@@ -134,7 +134,11 @@ typedef enum {
     /* Note: if you're hunting for a flag, some day in the future when we
      * have used them all, this one is easy enough to eliminate by having the
      * tiny number of objects marked this way in a remembered set. */
-    MVM_CF_NEVER_REPOSSESS = 2048
+    MVM_CF_NEVER_REPOSSESS = 2048,
+
+    /* Has this item been chained into a gen2 freelist? This is only used in
+     * GC debug more. */
+    MVM_CF_DEBUG_IN_GEN2_FREE_LIST = 4096,
 } MVMCollectableFlags;
 
 #ifdef MVM_USE_OVERFLOW_SERIALIZATION_INDEX
@@ -234,7 +238,7 @@ struct MVMBoolificationSpec {
 /* Constant for incrementing the type cache ID for new STables. This leaves
  * the lowest bits free for caches to attach flags (of note, the multi
  * dispatch cache). */
-#define MVM_TYPE_CACHE_ID_INCR 128
+#define MVM_TYPE_CACHE_ID_INCR 256
 
 /* S-table, representing a meta-object/representation pairing. Note that the
  * items are grouped in hope that it will pack decently and do decently in
@@ -349,6 +353,10 @@ struct MVMSTable {
     /* A string associated with this STable for debugging purposes.
      * Usually the name of the class this belongs to. */
     char *debug_name;
+
+    /* If this STable is currently in the process of being repossessed. Used
+     * to trigger clearup of memory pre-repossession. */
+    MVMuint8 being_repossessed;
 };
 
 /* The representation operations table. Note that representations are not
@@ -378,6 +386,13 @@ struct MVMREPROps_Attribute {
     MVMint64 (*is_attribute_initialized) (MVMThreadContext *tc, MVMSTable *st,
         void *data, MVMObject *class_handle, MVMString *name,
         MVMint64 hint);
+
+    /* Provided the attribute is a native integer of the architecture's atomic
+     * size, returns an AO_t * referencing it. This is only valid until the
+     * next safepoint. If rebless is called on the object, updates may be lost
+     * although memory safety must not be violated. */
+    AO_t * (*attribute_as_atomic) (MVMThreadContext *tc, MVMSTable *st,
+        void *data, MVMObject *class_handle, MVMString *name);
 };
 struct MVMREPROps_Boxing {
     /* Used with boxing. Sets an integer value, for representations that
@@ -494,6 +509,16 @@ struct MVMREPROps_Positional {
 
     /* Gets the STable representing the declared element type. */
     MVMStorageSpec (*get_elem_storage_spec) (MVMThreadContext *tc, MVMSTable *st);
+
+    /* Provided the array consists of integers of the architecture's atomic
+     * size, gets an AO_t * pointing to that element and valid until the next
+     * safepoint. */
+    AO_t * (*pos_as_atomic) (MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
+        void *data, MVMint64 index);
+
+    /* Multi-dim version of as_atomic. */
+    AO_t * (*pos_as_atomic_multidim) (MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
+        void *data, MVMint64 num_indices, MVMint64 *indices);
 };
 struct MVMREPROps_Associative {
     /* Gets the value at the specified key and places it in the passed
@@ -641,7 +666,8 @@ struct MVMREPROps {
 /* Some functions related to 6model core functionality. */
 MVM_PUBLIC MVMObject * MVM_6model_get_how(MVMThreadContext *tc, MVMSTable *st);
 MVM_PUBLIC MVMObject * MVM_6model_get_how_obj(MVMThreadContext *tc, MVMObject *obj);
-void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name, MVMRegister *res);
+void MVM_6model_find_method(MVMThreadContext *tc, MVMObject *obj, MVMString *name,
+    MVMRegister *res, MVMint64 throw_if_not_found);
 MVM_PUBLIC MVMObject * MVM_6model_find_method_cache_only(MVMThreadContext *tc, MVMObject *obj, MVMString *name);
 MVMint32 MVM_6model_find_method_spesh(MVMThreadContext *tc, MVMObject *obj, MVMString *name,
                                       MVMint32 ss_idx, MVMRegister *res);
@@ -654,3 +680,10 @@ void MVM_6model_invoke_default(MVMThreadContext *tc, MVMObject *invokee, MVMCall
 void MVM_6model_stable_gc_free(MVMThreadContext *tc, MVMSTable *st);
 MVMuint64 MVM_6model_next_type_cache_id(MVMThreadContext *tc);
 void MVM_6model_never_repossess(MVMThreadContext *tc, MVMObject *obj);
+
+MVM_STATIC_INLINE char *MVM_6model_get_debug_name(MVMThreadContext *tc, MVMObject *obj) {
+    return STABLE(obj)->debug_name ? STABLE(obj)->debug_name : "";
+}
+MVM_STATIC_INLINE char *MVM_6model_get_stable_debug_name(MVMThreadContext *tc, MVMSTable *stable) {
+    return stable->debug_name ? stable->debug_name : "";
+}
