@@ -21,6 +21,7 @@
 /* Frame flags. */
 #define FRAME_FLAG_EXIT_HANDLER     1
 #define FRAME_FLAG_IS_THUNK         2
+#define FRAME_FLAG_NO_INLINE        8
 
 /* Describes the current reader state. */
 typedef struct {
@@ -281,7 +282,7 @@ static void deserialize_sc_deps(MVMThreadContext *tc, MVMCompUnit *cu, ReaderSta
         handle = MVM_cu_string(tc, cu, sh_idx);
 
         /* See if we can resolve it. */
-        uv_mutex_lock(&tc->instance->mutex_sc_weakhash);
+        uv_mutex_lock(&tc->instance->mutex_sc_registry);
         MVM_HASH_GET(tc, tc->instance->sc_weakhash, handle, scb);
         if (scb && scb->sc) {
             cu_body->scs_to_resolve[i] = NULL;
@@ -298,7 +299,7 @@ static void deserialize_sc_deps(MVMThreadContext *tc, MVMCompUnit *cu, ReaderSta
             cu_body->scs_to_resolve[i] = scb;
             cu_body->scs[i] = NULL;
         }
-        uv_mutex_unlock(&tc->instance->mutex_sc_weakhash);
+        uv_mutex_unlock(&tc->instance->mutex_sc_registry);
     }
 }
 
@@ -516,6 +517,7 @@ static MVMStaticFrame ** deserialize_frames(MVMThreadContext *tc, MVMCompUnit *c
             MVMint16 flags = read_int16(pos, 38);
             static_frame_body->has_exit_handler = flags & FRAME_FLAG_EXIT_HANDLER;
             static_frame_body->is_thunk         = flags & FRAME_FLAG_IS_THUNK;
+            static_frame_body->no_inline        = flags & FRAME_FLAG_NO_INLINE;
         }
 
         /* Read code object SC indexes (version 4 and higher). */
@@ -582,7 +584,11 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
         return;
 
     /* Acquire the update mutex on the CompUnit. */
-    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
+    MVMROOT(tc, cu, {
+    MVMROOT(tc, sf, {
+        MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
+    });
+    });
 
     /* Ensure no other thread has done this for us in the mean time. */
     if (sf->body.fully_deserialized) {
@@ -648,7 +654,7 @@ void MVM_bytecode_finish_frame(MVMThreadContext *tc, MVMCompUnit *cu,
                 sf->body.handlers[j].label_reg = read_int16(pos, 0);
                 pos += 2;
             }
-            sf->body.handlers[j].inlined_and_not_lexical = 0;
+            sf->body.handlers[j].inlinee = -1;
         }
     }
 
@@ -923,7 +929,29 @@ MVMBytecodeAnnotation * MVM_bytecode_resolve_annotation(MVMThreadContext *tc, MV
         ba->bytecode_offset = read_int32(cur_anno, 0);
         ba->filename_string_heap_index = read_int32(cur_anno, 4);
         ba->line_number = read_int32(cur_anno, 8);
+        ba->ann_offset = cur_anno - sfb->annotations_data;
+        ba->ann_index  = i;
     }
 
     return ba;
 }
+
+void MVM_bytecode_advance_annotation(MVMThreadContext *tc, MVMStaticFrameBody *sfb, MVMBytecodeAnnotation *ba) {
+    MVMuint32 i = ba->ann_index + 1;
+
+    MVMuint8 *cur_anno = sfb->annotations_data + ba->ann_offset;
+    cur_anno += 12;
+    if (i >= sfb->num_annotations) {
+        ba->bytecode_offset = -1;
+        ba->filename_string_heap_index = 0;
+        ba->line_number = 0;
+        ba->ann_offset = -1;
+        ba->ann_index = -1;
+    } else {
+        ba->bytecode_offset = read_int32(cur_anno, 0);
+        ba->filename_string_heap_index = read_int32(cur_anno, 4);
+        ba->line_number = read_int32(cur_anno, 8);
+        ba->ann_offset = cur_anno - sfb->annotations_data;
+        ba->ann_index  = i;
+    }
+}
